/*
 *	Brainfuck interpreter
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BF_ARRAY_INIT_SIZE  256
#define BF_SCRIPT_INIT_SIZE 1024

#define err(s) ((void)(puts("Error: " s), abort()))

typedef char bf_t;

typedef struct
{
	bf_t *mem;
	bf_t *ptr;
	bf_t *end;
} bf_array;

bf_t *bf_array_create(bf_array *pary)
{
	bf_t *mem;
	
	mem = (bf_t *)malloc(sizeof(bf_t) * BF_ARRAY_INIT_SIZE);
	if (!mem)
		err("memory not enough");
	memset(mem, 0, sizeof(bf_t) * BF_ARRAY_INIT_SIZE);
	
	pary->mem = mem;
	pary->ptr = mem + BF_ARRAY_INIT_SIZE / 2;
	pary->end = mem + BF_ARRAY_INIT_SIZE;
	
	return pary->ptr;
}

void bf_array_destroy(bf_array *pary)
{
	free(pary->mem);
	
	return;
}

bf_t *bf_array_inc_ptr(bf_array *pary)
{
	size_t size;
	bf_t *newmem;
	
	if (++pary->ptr < pary->end)
		return pary->ptr;
	
	size = pary->end - pary->mem;
	newmem = (bf_t *)realloc(pary->mem, sizeof(bf_t) * size * 2);
	if (!newmem)
		err("memory not enough");
	memset(newmem + size, 0, sizeof(bf_t) * size);
	
	pary->mem = newmem;
	pary->ptr = newmem + size;
	pary->end = newmem + size * 2;
	
	return pary->ptr;
}

bf_t *bf_array_dec_ptr(bf_array *pary)
{
	size_t size;
	bf_t *newmem;
	
	if (pary->mem <= --pary->ptr)
		return pary->ptr;
	
	size = pary->end - pary->mem;
	newmem = (bf_t *)malloc(sizeof(bf_t) * size * 2);
	if (!newmem)
		err("memory not enough");
	memset(newmem, 0, sizeof(bf_t) * size);
	memcpy(newmem + size, pary->mem, sizeof(bf_t) * size);
	free(pary->mem);
	
	pary->mem = newmem;
	pary->ptr = newmem + size - 1;
	pary->end = newmem + size * 2;
	
	return pary->ptr;
}

typedef struct
{
	char *mem;
	char *ptr;
	char *end;
} bf_script;

void bf_script_create(bf_script *pscr)
{
	char *mem;
	
	mem = (char *)malloc(sizeof(char) * BF_SCRIPT_INIT_SIZE);
	if (!mem)
		err("memory not enough");
	
	pscr->mem = mem;
	pscr->ptr = mem;
	pscr->end = mem + BF_SCRIPT_INIT_SIZE;
	
	return;
}

void bf_script_destroy(bf_script *pscr)
{
	free(pscr->mem);
	
	return;
}

void bf_script_put(bf_script *pscr, int c)
{
	size_t size;
	char *newmem;
	
	if (!strchr("<>+-.,[]", c))
		return;
	
	if (pscr->end <= pscr->ptr) {
		size = pscr->end - pscr->mem;
		newmem = (char *)realloc(pscr->mem, sizeof(char) * size * 2);
		if (!newmem)
			err("memory not enough");
		
		pscr->mem = newmem;
		pscr->ptr = newmem + size;
		pscr->end = newmem + size * 2;
	}
	
	*pscr->ptr = (char)c;
	++pscr->ptr;
	
	return;
}

void bf_script_run(bf_script *pscr)
{
	bf_array ary;
	bf_t *ptr;
	char *s;
	
	ptr = bf_array_create(&ary);
	
	for (s = pscr->mem; s < pscr->end; ++s) {
		switch (*s) {
		case '<':
			ptr = bf_array_dec_ptr(&ary);
			break;
		case '>':
			ptr = bf_array_inc_ptr(&ary);
			break;
		case '+':
			++(*ptr);
			break;
		case '-':
			--(*ptr);
			break;
		case '.':
			putchar(*ptr);
			break;
		case ',':
			*ptr = (bf_t)getchar();
			break;
		case ']':
			do {
				if (--s < pscr->mem)
					err("'[' not found");
			} while (*s != '[');
			/* fall through */
		case '[':
			if (*ptr == 0) {
				do {
					if (++s >= pscr->end)
						err("']' not found");
				} while (*s != ']');
			}
			break;
		}
	}
	
	bf_array_destroy(&ary);
	
	return;
}

int main(int argc, char *argv[])
{
	bf_script scr;
	int c, c_prev;
	int i;
	FILE *fp;
	char *fn;
	
	bf_script_create(&scr);
	
	if (argc < 2) {
		puts("F*ck your brain ? (hit RETURN twice to run)");
		
		c_prev = EOF;
		while ((c = getchar()) != EOF) {
			if (c == '\n' && c_prev == '\n')
				break;
			bf_script_put(&scr, c);
			c_prev = c;
		}
	}
	for (i = 1; i < argc; ++i) {
		if ((fp = fopen(argv[i], "rb")) == NULL) {
			printf("Error: can't open file \"%s\"\n", argv[i]);
			abort();
		}
		while ((c = fgetc(fp)) != EOF)
			bf_script_put(&scr, c);
		fclose(fp);
	}
	
	bf_script_run(&scr);
	
	bf_script_destroy(&scr);
	
	return EXIT_SUCCESS;
}
